//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Data/DatafilesSelector.cpp
//! @brief     Implements class DatafilesSelector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Data/DatafilesSelector.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/IO/IOFactory.h"
#include "GUI/Model/Data/DataItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Files/DatafilesQModel.h"
#include "GUI/View/Data/DatafileEditor.h"
#include "GUI/View/Data/StackedFrames.h"
#include "GUI/View/Info/MessageBox.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Loader/Legacy1dDialog.h"
#include "GUI/View/Manager/ProjectManager.h"
#include "GUI/View/Widget/ItemViewOverlayButtons.h"
#include <QFileDialog>
#include <QMenu>
#include <QSplitter>
#include <QVBoxLayout>

namespace {

// The following two functions are templated on T in {IO::Filetype1D, IO::Filetype2D}.

template <class T>
QString join_filterkeys(std::vector<std::pair<const QString, T>> _nomap, const QString& _separator)
{
    QString result;
    for (const auto& it : _nomap) {
        if (!result.isEmpty())
            result += _separator;
        result += it.first;
    }
    return result;
}

template <class T>
T filterkey2type(std::vector<std::pair<const QString, T>> _nomap, const QString& _key)
{
    for (const auto& it : _nomap)
        if (it.first == _key)
            return it.second;
    ASSERT_NEVER;
}

} // namespace


DatafilesSelector::DatafilesSelector(StackedDataFrames* data_frames)
    : m_import1dData_action(new QAction(this))
    , m_import2dData_action(new QAction(this))
    , m_rename_data_action(new QAction(this))
    , m_remove_data_action(new QAction(this))
    , m_tree_view(new QTreeView(this))
    , m_tree_model(new DatafilesQModel(this, gDoc->realModel()))
    , m_editor(new DatafileEditor)
    , m_data_frames(data_frames)
{
    setMinimumSize(250, 600);
    setSizePolicy(QSizePolicy::Preferred, QSizePolicy::Expanding);
    setWindowTitle("DatafilesSelector");

    m_import2dData_action->setText("Import 2D data");
    m_import2dData_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import2dData_action->setIconText("2D");
    m_import2dData_action->setToolTip("Import 2D data");
    connect(m_import2dData_action, &QAction::triggered, [this] { importData2D(); });

    m_import1dData_action->setText("Import 1D data");
    m_import1dData_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_import1dData_action->setIconText("1D");
    m_import1dData_action->setToolTip("Import 1D data");
    connect(m_import1dData_action, &QAction::triggered, [this] { importData1D(); });

    m_rename_data_action->setText("Rename");
    m_rename_data_action->setIcon(QIcon()); // #baTODO: Icon needed?
    m_rename_data_action->setIconText("Rename");
    m_rename_data_action->setToolTip("Rename data");
    connect(m_rename_data_action, &QAction::triggered, this, &DatafilesSelector::renameCurrentItem);

    m_remove_data_action->setText("Remove");
    m_remove_data_action->setIcon(QIcon(":/images/delete.svg"));
    m_remove_data_action->setIconText("Remove");
    m_remove_data_action->setToolTip("Remove selected data");
    connect(m_remove_data_action, &QAction::triggered, this, &DatafilesSelector::removeCurrentItem);

    // m_tree_view->setAttribute(Qt::WA_MacShowFocusRect, false); // TODO/Mac: needed?
    m_tree_view->setItemsExpandable(false);
    m_tree_view->setRootIsDecorated(false);
    m_tree_view->setHeaderHidden(true);
    m_tree_view->setContextMenuPolicy(Qt::CustomContextMenu);
    m_tree_view->setModel(m_tree_model);

    auto* splitter = new QSplitter;
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_tree_view);
    splitter->addWidget(m_editor);
    splitter->setChildrenCollapsible(true);

    auto* mainLayout = new QVBoxLayout;
    mainLayout->setContentsMargins(0, 0, 0, 0);
    mainLayout->setSpacing(0);
    mainLayout->addWidget(splitter);

    setLayout(mainLayout);

    connect(m_editor, &DatafileEditor::instrumentUpdated, [this] {
        if (m_data_frames && currentDatafileItem())
            m_data_frames->setStackedItem(currentDatafileItem()->dataItem());
    });
    // TODO rm ??

    connect(m_tree_view, &QTreeView::customContextMenuRequested, this,
            &DatafilesSelector::onContextMenuRequest);

    connect(m_tree_model, &QAbstractItemModel::modelReset, [this] { m_tree_view->expandAll(); });

    connect(m_tree_model, &QAbstractItemModel::rowsInserted, [this] { m_tree_view->expandAll(); });

    connect(m_tree_view->selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &DatafilesSelector::onSelectionChanged);

    m_tree_view->expandAll();
    updateActionEnabling();

    restoreSelection();

    ItemViewOverlayButtons::install(
        m_tree_view, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
}

QSize DatafilesSelector::sizeHint() const
{
    return {200, 400};
}

QSize DatafilesSelector::minimumSizeHint() const
{
    return {128, 200};
}

DatafileItem* DatafilesSelector::currentDatafileItem()
{
    return m_tree_model->itemForIndex(currentDatafileIndex());
}

void DatafilesSelector::setCurrentDatafileItem(DatafileItem* item)
{
    m_tree_view->selectionModel()->clearSelection();
    QModelIndex index = m_tree_model->indexForItem(item);
    if (index.isValid())
        m_tree_view->selectionModel()->setCurrentIndex(index, QItemSelectionModel::SelectCurrent);
}

void DatafilesSelector::restoreSelection()
{
    int lastIndex = gDoc->realModel()->selectedIndex();
    int lastRank = gDoc->realModel()->selectedRank();
    QModelIndex lastUsedIndex =
        m_tree_model->index(lastIndex, 0, m_tree_model->indexOfHeadline(lastRank));

    if (lastUsedIndex.isValid())
        m_tree_view->selectionModel()->setCurrentIndex(lastUsedIndex,
                                                       QItemSelectionModel::SelectCurrent);
    else
        setCurrentDatafileItem(m_tree_model->topMostItem());

    if (currentDatafileItem())
        m_data_frames->setStackedItem(currentDatafileItem()->dataItem());
}

QModelIndex DatafilesSelector::currentDatafileIndex() const
{
    if (!m_tree_view->selectionModel()->isSelected(m_tree_view->selectionModel()->currentIndex()))
        return {};
    return m_tree_view->selectionModel()->currentIndex();
}

QList<QAction*> DatafilesSelector::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (m_tree_model->isHeadline(index)) {
        if (asHover)
            return {};
        return (index == m_tree_model->indexOfHeadline(1))
                   ? QList<QAction*>({m_import1dData_action})
                   : QList<QAction*>({m_import2dData_action});
    }

    // -- index belongs to item
    if (!asHover)
        return {};

    auto* item = m_tree_model->itemForIndex(index);
    if (item == nullptr)
        return {};

    auto* removeAction = new QAction(this);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this data");
    connect(removeAction, &QAction::triggered, [this, item] {
        m_tree_model->removeDataItem(item);
        gDoc->setModified();
    });

    return {removeAction};
}

void DatafilesSelector::onSelectionChanged()
{
    updateActionEnabling();

    gDoc->realModel()->setSelectedIndex(currentDatafileIndex().row());
    gDoc->realModel()->setSelectedRank(currentDatafileIndex().parent().row() + 1);

    m_editor->setDatafileItem(currentDatafileItem());
    m_data_frames->setStackedItem(currentDatafileItem()->dataItem());
}

void DatafilesSelector::onContextMenuRequest(const QPoint& point)
{
    auto* dfile_item = m_tree_model->itemForIndex(m_tree_view->indexAt(point));
    updateActionEnabling(dfile_item);

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (dfile_item != nullptr) {
        menu.addAction(m_rename_data_action);
        menu.addAction(m_remove_data_action);
        menu.addSeparator();
    }

    menu.addAction(m_import2dData_action);
    menu.addAction(m_import1dData_action);
    menu.exec(m_tree_view->mapToGlobal(point));
}

void DatafilesSelector::importData1D()
{
    static const std::vector<std::pair<const QString, IO::Filetype1D>> filters1D{
        {"Legacy table / interactive configuration (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv1D},
        {"Motofit (*.mft)", IO::mft},
        {"BornAgain (*.int *.int.gz)", IO::bornagain1D},
        {"all (*.*)", IO::unknown1D}};

    static const QString filters = ::join_filterkeys(filters1D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter1D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter1D(selectedFilter);

    const IO::Filetype1D global_ftype = ::filterkey2type(filters1D, selectedFilter);

    for (const QString& fname : fnames) {
        DatafileItem* dfile_item = nullptr;
        const ImportSettings1D* settings = nullptr;

        IO::Filetype1D ftype = global_ftype;
        if (ftype == IO::unknown1D)
            ftype = IO::filename2type1D(fname.toStdString());
        if (ftype == IO::csv1D) {
            Legacy1dDialog dialog(this, fname);
            const auto result = dialog.exec();
            if (result != QDialog::Accepted)
                return;
            settings = &Legacy1dDialog::Msettings;
        }

        try {
            std::unique_ptr<Datafield> df(IO::readData1D(fname.toStdString(), ftype, settings));
            if (!df)
                continue;
            if (df->rank() != 1)
                throw std::runtime_error("File does not contain a 1d data set");
            dfile_item = m_tree_model->injectDataItem(fname, *df);
            setCurrentDatafileItem(dfile_item);
        } catch (std::exception& ex) {
            m_tree_model->removeDataItem(dfile_item);

            const QString message = QString("Cannot read file %1:\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
        }
    }

    gDoc->setModified();
}

void DatafilesSelector::importData2D()
{
    static const std::vector<std::pair<const QString, IO::Filetype2D>> filters2D{
        {"all (*.*)", IO::unknown2D},
        {"TIFF (*.tif *.tiff *.tif.gz)", IO::tiff},
        {"Nicos/SANSDRaw (*.001)", IO::nicos2D},
        {"BornAgain (*.int *.int.gz)", IO::bornagain2D},
        {"CSV (*.csv *.dat *.tab *.txt *.dat.gz)", IO::csv2D},
        {"REFSANS (*.csv)", IO::refsans2D}};

    static const QString filters = ::join_filterkeys(filters2D, ";;");

    QString selectedFilter = ProjectManager::instance()->recentlyUsedImportFilter2D();
    const QString dirname = ProjectManager::instance()->userImportDir();

    const QStringList fnames = QFileDialog::getOpenFileNames(
        Q_NULLPTR, "Open Intensity Files", dirname, filters, &selectedFilter,
        appSettings->useNativeFileDialog() ? QFileDialog::Options()
                                           : QFileDialog::DontUseNativeDialog);
    if (fnames.isEmpty())
        return;

    ProjectManager::instance()->setImportDirFromFilePath(fnames[0]);
    ProjectManager::instance()->setRecentlyUsedImportFilter2D(selectedFilter);

    const IO::Filetype2D global_ftype = ::filterkey2type(filters2D, selectedFilter);

    for (const QString& fname : fnames) {
        IO::Filetype2D ftype = global_ftype;
        if (ftype == IO::unknown2D)
            ftype = IO::filename2type2D(fname.toStdString());

        DatafileItem* dfile_item = nullptr;
        try {
            std::unique_ptr<Datafield> df(IO::readData2D(fname.toStdString(), ftype));
            if (!df)
                continue;
            if (df->rank() != 2)
                throw std::runtime_error("File does not contain a 2d data set");
            dfile_item = m_tree_model->injectDataItem(fname, *df);
            setCurrentDatafileItem(dfile_item);
        } catch (std::exception& ex) {
            m_tree_model->removeDataItem(dfile_item);

            const QString message = QString("Cannot read file %1\n\n%2")
                                        .arg(fname)
                                        .arg(QString::fromStdString(std::string(ex.what())));
            GUI::Message::warning(this, "File import", message);
        }
    }

    gDoc->setModified();
}

void DatafilesSelector::renameCurrentItem()
{
    if (currentDatafileIndex().isValid()) {
        // First delete any existing index widget. This is necessary if overlay buttons are used,
        // because these buttons use the indexWidget() functionality, which itself uses the same
        // internals as an item editor
        m_tree_view->setIndexWidget(currentDatafileIndex(), nullptr);
        m_tree_view->edit(currentDatafileIndex());
        gDoc->setModified();
    }
}

void DatafilesSelector::removeCurrentItem()
{
    m_tree_model->removeDataItem(currentDatafileItem());
    gDoc->setModified();
}

void DatafilesSelector::updateActionEnabling()
{
    updateActionEnabling(currentDatafileItem());
}

void DatafilesSelector::updateActionEnabling(const DatafileItem* item) const
{
    m_remove_data_action->setEnabled(item != nullptr);
    m_rename_data_action->setEnabled(item != nullptr);
}
