//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentLibraryDialog.cpp
//! @brief     Implements class InstrumentLibraryDialog.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentLibraryDialog.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentLibrary.h"
#include "GUI/Support/Style/Style.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Layout/mainwindow_constants.h"
#include "GUI/View/Tool/ItemDelegateForHTML.h"
#include "GUI/View/Widget/ItemViewOverlayButtons.h"
#include <QAction>
#include <QFormLayout>
#include <QInputDialog>
#include <QPushButton>
#include <QSettings>
#include <QTextEdit>
#include <QVBoxLayout>

InstrumentLibraryDialog::InstrumentLibraryDialog(QWidget* parent,
                                                 InstrumentLibrary* instrumentLibrary)
    : QDialog(parent)
    , m_instrument_library(instrumentLibrary)
    , m_tree_model(new InstrumentsQTreeModel(this, instrumentLibrary->instrumentModel()))
    , m_chosen_item(nullptr)
{
    // object name is needed to reload application settings
    if (this->objectName().isEmpty())
        this->setObjectName("InstrumentLibraryDialog");

    setGeometry(0, 0, 320, 429);

    auto* layout = new QVBoxLayout(this);
    setLayout(layout);

    m_tree_view = new QTreeView;
    layout->addWidget(m_tree_view);

    m_button_box = new QDialogButtonBox;
    layout->addWidget(m_button_box);
    m_button_box->setOrientation(Qt::Horizontal);
    m_button_box->setStandardButtons(QDialogButtonBox::Cancel | QDialogButtonBox::Ok);

    setWindowIcon(QIcon(":/images/library.svg"));
    setWindowFlag(Qt::WindowContextHelpButtonHint, false);

    m_tree_model->enableEmptyHeadlines(false);

    m_tree_view->setItemsExpandable(false);
    m_tree_view->setRootIsDecorated(false);
    m_tree_view->setHeaderHidden(true);
    m_tree_view->setContextMenuPolicy(Qt::CustomContextMenu);
    m_tree_view->setModel(m_tree_model);
    m_tree_view->expandAll();
    m_tree_view->setVerticalScrollMode(QTreeView::ScrollPerPixel);
    m_tree_view->setIndentation(0);
    m_tree_view->setItemDelegate(new ItemDelegateForHTML(this));
    m_tree_view->setIconSize(QSize(128, 128));

    connect(m_tree_model, &QAbstractItemModel::modelReset, [this] { m_tree_view->expandAll(); });

    connect(m_button_box, &QDialogButtonBox::accepted, this, &QDialog::accept);
    connect(m_button_box, &QDialogButtonBox::rejected, this, &QDialog::reject);

    // ensure a current item when widget is shown
    GUI::Style::setResizable(this);
    appSettings->loadWindowSizeAndPos(this);
}

InstrumentLibraryDialog::~InstrumentLibraryDialog()
{
    appSettings->saveWindowSizeAndPos(this);
}

InstrumentItem* InstrumentLibraryDialog::execChoose()
{
    setWindowTitle("Instrument Library - Choose instrument");

    ItemViewOverlayButtons::install(
        m_tree_view, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    m_tree_view->setItemDelegate(new ItemDelegateForHTML(this));

    connect(m_tree_view, &QTreeView::doubleClicked, this,
            &InstrumentLibraryDialog::onItemDoubleClickedForChoose);
    connect(m_tree_view->selectionModel(), &QItemSelectionModel::currentChanged, this,
            &InstrumentLibraryDialog::onCurrentChangedForChoose);
    onCurrentChangedForChoose();

    // select the first instrument
    QModelIndex index = m_tree_model->indexForItem(m_instrument_library->instrumentItems().first());
    m_tree_view->setCurrentIndex(index);

    if (exec() == QDialog::Accepted)
        return m_chosen_item;

    return nullptr;
}

void InstrumentLibraryDialog::execAdd(const InstrumentItem& instrumentToAdd)
{
    const QString& newName = m_instrument_library->suggestName(instrumentToAdd.instrumentName());
    auto* addedInstrument = m_instrument_library->addItemCopy(newName, instrumentToAdd);

    setWindowTitle("Instrument Library - Add instrument");

    m_tree_model->setNewInstrument(addedInstrument);

    ItemViewOverlayButtons::install(
        m_tree_view, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    m_tree_view->setItemDelegate(new ItemDelegateForHTML(this));

    QModelIndex index = m_tree_model->indexForItem(addedInstrument);
    m_tree_view->expandAll();
    m_tree_view->setCurrentIndex(index);
    m_tree_view->scrollTo(index, QAbstractItemView::PositionAtTop);
    exec();
}

void InstrumentLibraryDialog::onItemDoubleClickedForChoose(const QModelIndex& index)
{
    m_chosen_item = m_tree_model->itemForIndex(index);
    if (m_chosen_item != nullptr)
        accept();
}

void InstrumentLibraryDialog::onCurrentChangedForChoose()
{
    m_chosen_item = m_tree_model->itemForIndex(m_tree_view->currentIndex());
    m_button_box->button(QDialogButtonBox::Ok)->setEnabled(m_chosen_item != nullptr);
}

QList<QAction*> InstrumentLibraryDialog::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (m_tree_model->isHeadline(index))
        return {};

    // -- index belongs to item
    if (!asHover)
        return {};

    auto* item = m_tree_model->itemForIndex(index);
    if (item == nullptr)
        return {};

    auto* removeAction = new QAction(this);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this instrument");
    connect(removeAction, &QAction::triggered, [this, item] { m_tree_model->removeItem(item); });

    return {removeAction};
}

void InstrumentLibraryDialog::onInstrumentNameEdited(const QString& newName)
{
    QModelIndex index = m_tree_view->currentIndex();
    m_tree_model->setData(index, newName, Qt::EditRole);
}

void InstrumentLibraryDialog::onInstrumentDescriptionEdited(const QString& t)
{
    QModelIndex index = m_tree_view->currentIndex();
    m_tree_model->setData(index, t, Qt::ToolTipRole);
}
