//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentView.cpp
//! @brief     Implements class InstrumentView.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentView.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Instrument/DepthprobeInstrumentEditor.h"
#include "GUI/View/Instrument/GISASInstrumentEditor.h"
#include "GUI/View/Instrument/InstrumentListing.h"
#include "GUI/View/Instrument/OffspecInstrumentEditor.h"
#include "GUI/View/Instrument/SpecularInstrumentEditor.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Widget/GroupBoxes.h"
#include "GUI/View/Widget/StyledToolbar.h"
#include <QFormLayout>
#include <QLineEdit>
#include <QTextEdit>

InstrumentView::InstrumentView()
    : m_listing(new InstrumentListing)
    , m_scroll_area(new QScrollArea)
{
    auto* layout = new QVBoxLayout(this);

    // Top toolbar with action buttons "new ...instrument" etc
    auto* toolbar = new StyledToolbar;
    layout->addWidget(toolbar);
    toolbar->setToolButtonStyle(Qt::ToolButtonTextBesideIcon);

    // Everything below
    auto* hLayout = new QHBoxLayout;
    layout->addLayout(hLayout);

    // Left margin: instrument list

    hLayout->addWidget(m_listing);

    // Large widget: current instrument

    hLayout->addWidget(m_scroll_area);
    m_scroll_area->setWidgetResizable(true);
    m_scroll_area->setMinimumWidth(1000);

    hLayout->addStretch(1);

    //... Finalize

    toolbar->addActions(m_listing->toolbarActions());

    connect(m_listing, &InstrumentListing::instrumentSelected, this,
            &InstrumentView::createWidgetsForCurrentInstrument);

    // Initial state
    createWidgetsForCurrentInstrument();
    subscribeOnExternalUpdates();
}

void InstrumentView::subscribeOnExternalUpdates()
{
    connect(gDoc->multiNotifier(), &MultiInstrumentNotifier::instrumentChanged, this,
            &InstrumentView::onInstrumentChangedFromExternal, Qt::UniqueConnection);
}

void InstrumentView::showEvent(QShowEvent*)
{
    // disconnect because when this view is visible, no other instance is modifying instruments. By
    // disconnecting, no additional logic is necessary to avoid recursive calls (recursive since
    // this view also causes instrumentChanged to be emitted).
    disconnect(gDoc->multiNotifier(), &MultiInstrumentNotifier::instrumentChanged, this,
               &InstrumentView::onInstrumentChangedFromExternal);
}

void InstrumentView::hideEvent(QHideEvent*)
{
    // when the instrument view gets hidden (meaning another view is shown), it's necessary to
    // listen to changes (e.g. performed by the LinkInstrumentManager).
    subscribeOnExternalUpdates();
}

void InstrumentView::createWidgetsForCurrentInstrument()
{
    auto* currentInstrument = m_listing->currentInstrumentItem();
    if (!currentInstrument) {
        m_scroll_area->setWidget(new QWidget(m_scroll_area)); // blank widget
        return;
    }

    auto* w = new QWidget(m_scroll_area);
    auto* layout = new QVBoxLayout(w);

    //... Groupbox with instrument name and description (same layout for all instrument types)

    auto title = QString("Summary (%1 instrument)").arg(currentInstrument->instrumentType());
    auto* g = new CollapsibleGroupBox(title, m_scroll_area, currentInstrument->expandInfo);
    g->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Fixed);
    auto* formLayout = new QFormLayout;
    g->body()->setLayout(formLayout);
    layout->addWidget(g);

    auto* nameEdit = new QLineEdit(g);
    nameEdit->setText(currentInstrument->instrumentName());
    connect(nameEdit, &QLineEdit::textEdited, this, &InstrumentView::onInstrumentNameEdited);
    formLayout->addRow("Name:", nameEdit);

    auto* descriptionEdit = new QTextEdit(g);
    descriptionEdit->setMinimumWidth(300);
    descriptionEdit->setFixedHeight(60); // TODO replace by 2*line_height
    descriptionEdit->setAcceptRichText(false);
    descriptionEdit->setTabChangesFocus(true);
    descriptionEdit->setPlainText(currentInstrument->description());
    connect(descriptionEdit, &QTextEdit::textChanged, [this, descriptionEdit] {
        onInstrumentdescriptionEdited(descriptionEdit->toPlainText());
    });
    formLayout->addRow("Description:", descriptionEdit);

    //... All remaining content depends on instrument type

    IComponentEditor* editor;
    if (auto* ii = dynamic_cast<SpecularInstrumentItem*>(currentInstrument))
        editor = new SpecularInstrumentEditor(ii);
    else if (auto* ii = dynamic_cast<OffspecInstrumentItem*>(currentInstrument))
        editor = new OffspecInstrumentEditor(ii);
    else if (auto* ii = dynamic_cast<GISASInstrumentItem*>(currentInstrument))
        editor = new GISASInstrumentEditor(ii);
    else if (auto* ii = dynamic_cast<DepthprobeInstrumentItem*>(currentInstrument))
        editor = new DepthprobeInstrumentEditor(ii);
    else
        ASSERT_NEVER;

    layout->addWidget(editor);
    connect(editor, &IComponentEditor::dataChanged, this,
            &InstrumentView::onInstrumentChangedByEditor);

    m_scroll_area->setWidget(w);
}

void InstrumentView::onInstrumentNameEdited(const QString& newName)
{
    auto* currentInstrument = m_listing->currentInstrumentItem();
    if (currentInstrument && currentInstrument->instrumentName() != newName)
        gDoc->multiNotifier()->setInstrumentName(currentInstrument, newName);
}

void InstrumentView::onInstrumentdescriptionEdited(const QString& t)
{
    auto* currentInstrument = m_listing->currentInstrumentItem();
    if (currentInstrument && currentInstrument->description() != t) {
        currentInstrument->setDescription(t);
        onInstrumentChangedByEditor();
    }
}

void InstrumentView::onInstrumentChangedByEditor()
{
    // uses 'MultiInstrumentNotifier::instrumentChanged' signal for two purposes:
    // 1) notify 'ProjectDocument' that user has changed data ==> mark project with '*'
    // 2) notify 'LinkInstrumentManager' ==> unlink instrument from data if they are incompatible
    gDoc->multiNotifier()->notifyInstrumentChanged(m_listing->currentInstrumentItem());
}

void InstrumentView::onInstrumentChangedFromExternal(const InstrumentItem* instrument)
{
    if (instrument == m_listing->currentInstrumentItem())
        createWidgetsForCurrentInstrument();
}
