//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentsQListModel.cpp
//! @brief     Implements class InstrumentsQListModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentsQListModel.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/MultiInstrumentNotifier.h"
#include "GUI/Model/Project/ProjectDocument.h"

namespace {

/// Get the default name for an instrument (i.e. the name a newly created instrument should get)
template <typename T> QString defaultInstrumentName()
{
    return "Untitled";
}

template <> QString defaultInstrumentName<SpecularInstrumentItem>()
{
    return "Specular";
}

template <> QString defaultInstrumentName<DepthprobeInstrumentItem>()
{
    return "Depthprobe";
}

template <> QString defaultInstrumentName<GISASInstrumentItem>()
{
    return "GISAS";
}

template <> QString defaultInstrumentName<OffspecInstrumentItem>()
{
    return "Offspec";
}

} // namespace


InstrumentsQListModel::InstrumentsQListModel()
    : m_notifier(gDoc->multiNotifier())
{
    m_gisas_icon.addPixmap(QPixmap(":/images/gisas_instrument.svg"), QIcon::Selected);
    m_gisas_icon.addPixmap(QPixmap(":/images/gisas_instrument_shaded.svg"), QIcon::Normal);
    m_offspec_icon.addPixmap(QPixmap(":/images/offspec_instrument.svg"), QIcon::Selected);
    m_offspec_icon.addPixmap(QPixmap(":/images/offspec_instrument_shaded.svg"), QIcon::Normal);
    m_specular_icon.addPixmap(QPixmap(":/images/specular_instrument.svg"), QIcon::Selected);
    m_specular_icon.addPixmap(QPixmap(":/images/specular_instrument_shaded.svg"), QIcon::Normal);
    m_depth_probe_icon.addPixmap(QPixmap(":/images/depth_instrument.svg"), QIcon::Selected);
    m_depth_probe_icon.addPixmap(QPixmap(":/images/depth_instrument_shaded.svg"), QIcon::Normal);

    connect(m_notifier, &MultiInstrumentNotifier::instrumentNameChanged, this,
            &InstrumentsQListModel::onInstrumentNameChanged);
}

int InstrumentsQListModel::rowCount(const QModelIndex&) const
{
    return m_notifier->instrumentModel()->instrumentItems().size();
}

QVariant InstrumentsQListModel::data(const QModelIndex& index, int role) const
{
    QVector<InstrumentItem*> instruments = m_notifier->instrumentModel()->instrumentItems();
    if (!index.isValid() || index.row() >= instruments.size() || index.row() < 0)
        return {};

    const InstrumentItem* item = instruments[index.row()];
    if (role == Qt::DecorationRole) {
        if (dynamic_cast<const GISASInstrumentItem*>(item))
            return m_gisas_icon;
        if (dynamic_cast<const OffspecInstrumentItem*>(item))
            return m_offspec_icon;
        if (dynamic_cast<const SpecularInstrumentItem*>(item))
            return m_specular_icon;
        if (dynamic_cast<const DepthprobeInstrumentItem*>(item))
            return m_depth_probe_icon;
        return {};
    }

    if (role == Qt::DisplayRole)
        return item->instrumentName();

    return {};
}

InstrumentItem* InstrumentsQListModel::instrumentItemForIndex(const QModelIndex& index) const
{
    if (!index.isValid())
        return nullptr;

    QVector<InstrumentItem*> instruments = m_notifier->instrumentModel()->instrumentItems();
    if (index.row() >= 0 && index.row() < instruments.size())
        return instruments[index.row()];
    return nullptr;
}

QModelIndex InstrumentsQListModel::addNewGISASInstrument()
{
    return addNewInstrument<GISASInstrumentItem>();
}

QModelIndex InstrumentsQListModel::addNewOffspecInstrument()
{
    return addNewInstrument<OffspecInstrumentItem>();
}

QModelIndex InstrumentsQListModel::addNewSpecularInstrument()
{
    return addNewInstrument<SpecularInstrumentItem>();
}

QModelIndex InstrumentsQListModel::addNewDepthprobeInstrument()
{
    return addNewInstrument<DepthprobeInstrumentItem>();
}

void InstrumentsQListModel::removeInstrument(const QModelIndex& index)
{
    beginRemoveRows(QModelIndex(), index.row(), index.row());
    InstrumentItem* instrument = instrumentItemForIndex(index);
    m_notifier->removeInstrument(instrument);
    endRemoveRows();
}

QModelIndex InstrumentsQListModel::copyInstrument(const QModelIndex& source)
{
    const InstrumentItem* srcInstr = instrumentItemForIndex(source);
    ASSERT(srcInstr);

    return copyInstrument(srcInstr);
}

QModelIndex InstrumentsQListModel::copyInstrument(const InstrumentItem* source)
{
    const QString copyName =
        m_notifier->instrumentModel()->suggestInstrumentName(source->instrumentName());
    const int row = m_notifier->instrumentModel()->instrumentItems().size();

    beginInsertRows(QModelIndex(), row, row);
    m_notifier->addInstrumentItemCopy(source, copyName);
    endInsertRows();

    return createIndex(row, 0);
}

template <class Instrument> QModelIndex InstrumentsQListModel::addNewInstrument()
{
    const QString name =
        m_notifier->instrumentModel()->suggestInstrumentName(defaultInstrumentName<Instrument>());
    const int row = m_notifier->instrumentModel()->instrumentItems().size();

    beginInsertRows(QModelIndex(), row, row);
    auto* instrument = m_notifier->addInstrumentItem<Instrument>();
    m_notifier->setInstrumentName(instrument, name);
    endInsertRows();

    return createIndex(row, 0);
}

void InstrumentsQListModel::onInstrumentNameChanged(const InstrumentItem* instrument)
{
    const auto instruments = m_notifier->instrumentModel()->instrumentItems();
    if (const auto row = instruments.indexOf(instrument); row != -1)
        emit dataChanged(index(row, 0), index(row, 0));
}
