//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/JobControl/JobListModel.cpp
//! @brief     Implements class JobListModel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/JobControl/JobListModel.h"
#include "GUI/Model/Job/BatchInfo.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Project/ProjectDocument.h"

//==================================================================================================
// JobListModel
//==================================================================================================

//--------------------------------------------------------------------------------------------------
// public member functions
//--------------------------------------------------------------------------------------------------

JobListModel::JobListModel(JobsSet* jobs, QObject* parent)
    : QAbstractListModel(parent)
    , m_jobs(jobs)
{
    connect(jobs, &JobsSet::jobAdded, this, &JobListModel::onJobAdded);
    onJobAdded();
}

JobListModel::~JobListModel()
{
    for (JobItem* job : m_jobs->jobItems())
        disconnect(job, nullptr, this, nullptr);
}

int JobListModel::rowCount(const QModelIndex&) const
{
    return m_jobs->jobItems().size();
}

QVariant JobListModel::data(const QModelIndex& index, int role) const
{
    QVector<JobItem*> jobs = m_jobs->jobItems();
    if (!index.isValid() || index.row() >= jobs.size() || index.row() < 0)
        return {};

    JobItem* item = jobs[index.row()];
    if (role == Qt::DisplayRole)
        return item->batchInfo()->jobName();

    return {};
}

JobItem* JobListModel::jobItemForIndex(const QModelIndex& index) const
{
    QVector<JobItem*> jobs = m_jobs->jobItems();
    if (index.row() >= 0 && index.row() < jobs.size())
        return jobs[index.row()];
    return nullptr;
}

QModelIndex JobListModel::indexForJob(JobItem* job)
{
    QVector<JobItem*> jobs = m_jobs->jobItems();
    int idx = jobs.indexOf(job);
    if (idx != -1)
        return index(idx, 0);
    return {};
}

void JobListModel::runJob(const QModelIndex& index)
{
    m_jobs->runJob(jobItemForIndex(index));
}

void JobListModel::removeJob(const QModelIndex& index)
{
    beginRemoveRows(QModelIndex(), index.row(), index.row());
    JobItem* job = jobItemForIndex(index);
    disconnect(job, nullptr, this, nullptr);
    m_jobs->removeJob(job);
    endRemoveRows();
}

void JobListModel::cancelJob(const QModelIndex& index)
{
    m_jobs->cancelJob(jobItemForIndex(index));
}

//--------------------------------------------------------------------------------------------------
// private slots
//--------------------------------------------------------------------------------------------------

void JobListModel::emitJobListModelChanged(JobItem* job)
{
    QVector<JobItem*> jobs = m_jobs->jobItems();
    int i = jobs.indexOf(job);
    if (i != -1) {
        QModelIndex idx = index(i, 0);
        emit dataChanged(idx, idx);
    }
    gDoc->setModified();
}

void JobListModel::onJobAdded()
{
    for (JobItem* job : m_jobs->jobItems()) {
        connect(job->batchInfo(), &BatchInfo::jobStatusChanged,
                [this, job] { emitJobListModelChanged(job); });
        connect(job->batchInfo(), &BatchInfo::jobNameChanged,
                [this, job] { emitJobListModelChanged(job); });
        connect(job->batchInfo(), &BatchInfo::jobProgressChanged,
                [this, job] { emitJobListModelChanged(job); });
    }
}
