//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/JobControl/JobSelector.cpp
//! @brief     Implements class JobSelector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/JobControl/JobSelector.h"
#include "Base/Util/Assert.h"
#include "GUI/View/JobControl/JobListing.h"
#include "GUI/View/JobControl/JobPropertiesWidget.h"
#include "GUI/View/Layout/mainwindow_constants.h"
#include <QVBoxLayout>

namespace {

QVariant listToQVariant(const QList<int>& list)
{
    QList<QVariant> var_list;
    for (int val : list)
        var_list.push_back(QVariant(val));
    return {var_list};
}

QList<int> qVariantToList(const QVariant& var)
{
    QList<QVariant> var_list = var.toList();
    QList<int> list;
    for (const QVariant& var_val : var_list)
        list.push_back(var_val.toInt());
    return list;
}

} // namespace
JobSelector::JobSelector(JobsSet* jobModel, QWidget* parent)
    : QWidget(parent)
    , m_splitter(new QSplitter(Qt::Vertical, this))
    , m_job_properties(new JobPropertiesWidget(this))
    , m_job_model(jobModel)
{
    setWindowTitle(GUI::Style::JobSelectorName);
    setObjectName("JobSelector");

    auto* layout = new QVBoxLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);


    m_splitter->setChildrenCollapsible(true);
    layout->addWidget(m_splitter);

    m_listing = new JobListing(m_job_model, m_splitter);
    m_splitter->addWidget(m_listing);
    m_splitter->setCollapsible(0, false);
    m_splitter->setStretchFactor(0, 1);

    m_splitter->addWidget(m_job_properties);
    m_splitter->setCollapsible(1, false);
    m_splitter->setStretchFactor(1, 0);

    connect(m_listing, &JobListing::selectedJobsChanged, this, &JobSelector::onSelectedJobsChanged);

    onSelectedJobsChanged(selectedJobItems());

    setFixedWidth(GUI::Style::JOB_SELECTOR_PANEL_WIDTH);
    applySettings();
}

JobSelector::~JobSelector()
{
    saveSettings();
}

void JobSelector::resizeEvent(QResizeEvent* event)
{
    QWidget::resizeEvent(event);
    setMinimumWidth(0);
    setMaximumWidth(QWIDGETSIZE_MAX);
}

QVector<JobItem*> JobSelector::selectedJobItems() const
{
    return m_listing->selectedJobItems();
}

void JobSelector::makeJobItemSelected(JobItem* item)
{
    ASSERT(item);
    m_listing->selectJob(item);
}

void JobSelector::onSelectedJobsChanged(const QVector<JobItem*>& jobs)
{
    if (jobs.size() == 1)
        m_job_properties->setJobItem(jobs.front());
    else
        m_job_properties->setJobItem(nullptr);
    emit selectedJobsChanged();
}

void JobSelector::applySettings()
{
    QSettings settings;
    if (settings.childGroups().contains(GUI::Style::S_JOB_SELECTOR_PANEL)) {
        settings.beginGroup(GUI::Style::S_JOB_SELECTOR_PANEL);
        setFixedWidth(settings.value(GUI::Style::S_JOB_SELECTOR_PANEL_WIDTH).toInt());
        m_splitter->setSizes(
            qVariantToList(settings.value(GUI::Style::S_JOB_SELECTOR_SPLITTER_SIZES)));
        settings.endGroup();
    }
}

void JobSelector::saveSettings()
{
    QSettings settings;
    settings.beginGroup(GUI::Style::S_JOB_SELECTOR_PANEL);
    settings.setValue(GUI::Style::S_JOB_SELECTOR_PANEL_WIDTH, width());
    settings.setValue(GUI::Style::S_JOB_SELECTOR_SPLITTER_SIZES,
                      listToQVariant(m_splitter->sizes()));
    settings.endGroup();
    settings.sync();
}
