//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Setup/MaskPanel.cpp
//! @brief     Implements class MaskPanel.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Setup/MaskPanel.h"
#include "Fit/Param/RealLimits.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MasksQModel.h"
#include "GUI/Model/Mask/MasksSet.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/View/Numeric/DoubleSpinBox.h"
#include "GUI/View/Tool/LayoutUtil.h"
#include "GUI/View/Widget/GroupBoxes.h"
#include <QCheckBox>
#include <QLineEdit>

namespace {

void addSpinBox(MaskItem* mask, QFormLayout* layout, DoubleProperty& property)
{
    auto* spinbox = new DoubleSpinBox(&property);
    layout->addRow(property.label() + ":", spinbox);
    QObject::connect(spinbox, &DoubleSpinBox::valueChanged, [mask](double) {
        gDoc->setModified();
        ASSERT(mask);
        emit mask->maskGeometryChanged();
    });
    QObject::connect(&property, &DoubleProperty::setAndNotifyCalled, spinbox,
                     &DoubleSpinBox::updateValue);
}

} // namespace


MaskPanel::MaskPanel()
    : m_list_view(new QListView)
    , m_mask_list_model(nullptr)
    , m_current_mask_item(nullptr)
{
    setSizePolicy(QSizePolicy::MinimumExpanding, QSizePolicy::Expanding);
    setObjectName(QLatin1String("MaskEditorToolPanel"));

    m_list_view->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(m_list_view, &QListView::customContextMenuRequested, this,
            &MaskPanel::onCustomContextMenuRequested);

    auto* layout = new QVBoxLayout(this);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(8);

    // -- mask stack
    auto* g2 = new StaticGroupBox("Mask stack", this);
    layout->addWidget(g2);
    auto* l2 = new QVBoxLayout(g2->body());
    l2->setContentsMargins(0, 0, 0, 0);
    m_list_view->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    l2->addWidget(m_list_view);

    // -- mask properties
    auto* g3 = new StaticGroupBox("Selected mask", this);
    layout->addWidget(g3);
    g3->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    m_editor_layout = new QFormLayout(g3->body());
    m_editor_layout->setContentsMargins(0, 0, 0, 0);
    // contents of m_editor_layout is set by function setCurrentMaskItem
}

QSize MaskPanel::sizeHint() const
{
    return {128, 128};
}

QSize MaskPanel::minimumSizeHint() const
{
    return {128, 128};
}

void MaskPanel::updateMaskPanel(MasksQModel* mask_list_model)
{
    m_mask_list_model = mask_list_model;

    m_list_view->setModel(m_mask_list_model);
    m_list_view->setSelectionMode(QAbstractItemView::ExtendedSelection);
    m_list_view->setRootIndex(m_mask_list_model->maskContItem()->rootIndex);

    connect(m_list_view->selectionModel(), &QItemSelectionModel::selectionChanged, this,
            &MaskPanel::onSelectionChanged);
}

void MaskPanel::resetMaskPanel()
{
    m_mask_list_model = nullptr;
    m_list_view->setModel(nullptr);
    setCurrentMaskItem(nullptr);
}

QItemSelectionModel* MaskPanel::selectionModel()
{
    ASSERT(m_list_view);
    return m_list_view->selectionModel();
}

//! Show/Hide panel. When panel is hidden, all property editors are disabled.
void MaskPanel::setPanelHidden(bool hidden)
{
    setHidden(hidden);

    if (hidden) {
        setCurrentMaskItem(nullptr);
    } else {
        QModelIndexList indexes = selectionModel()->selectedIndexes();
        if (!indexes.empty())
            setCurrentMaskItem(m_mask_list_model->itemForIndex(indexes.front()));
    }
}

void MaskPanel::onSelectionChanged(const QItemSelection& selected, const QItemSelection&)
{
    if (m_keep_selection)
        return;

    if (!selected.empty())
        setCurrentMaskItem(m_mask_list_model->itemForIndex(selected.indexes().front()));
    else
        setCurrentMaskItem(nullptr);
}

void MaskPanel::onCustomContextMenuRequested(const QPoint& point)
{
    emit itemContextMenuRequest(m_list_view->mapToGlobal(point));
}

//! Private function, called by different functions.
//! Overwrites m_editor_layout.
void MaskPanel::setCurrentMaskItem(MaskItem* maskItem)
{
    GUI::Util::Layout::clearLayout(m_editor_layout);

    m_current_mask_item = maskItem;
    if (!m_current_mask_item)
        return;

    // -- mask value (only if not RoI)
    if (!dynamic_cast<RegionOfInterestItem*>(maskItem)) {
        const auto maskValueGetter = [maskItem] { return maskItem->maskValue(); };
        const auto maskValueSetter = [maskItem](bool b) { maskItem->setMaskValue(b); };
        addMaskCheckBox("Mask value", maskValueGetter, maskValueSetter);
    }
    // -- mask visibility
    const auto visibilityValueGetter = [maskItem] { return maskItem->isVisible(); };
    const auto visibilityValueSetter = [this, maskItem](bool b) {
        m_keep_selection = true;
        maskItem->setIsVisible(b);
        maskItem->setWasVisible(b);
        m_keep_selection = false;
    };
    addMaskCheckBox("Show", visibilityValueGetter, visibilityValueSetter);

    // -- name (only if not RoI)
    if (!dynamic_cast<RegionOfInterestItem*>(maskItem)) {
        auto* edit = new QLineEdit(maskItem->maskName(), m_editor_layout->parentWidget());
        connect(edit, &QLineEdit::textEdited,
                [maskItem](const QString& t) { maskItem->setMaskName(t); });
        connect(edit, &QLineEdit::editingFinished, [] { gDoc->setModified(); });
        m_editor_layout->addRow("Name:", edit);
    }

    if (auto* c = dynamic_cast<RectangleItem*>(maskItem)) {
        addSpinBox(maskItem, m_editor_layout, c->xLow());
        addSpinBox(maskItem, m_editor_layout, c->xUp());
        addSpinBox(maskItem, m_editor_layout, c->yLow());
        addSpinBox(maskItem, m_editor_layout, c->yUp());
    } else if (auto* c = dynamic_cast<EllipseItem*>(maskItem)) {
        addSpinBox(maskItem, m_editor_layout, c->xCenter());
        addSpinBox(maskItem, m_editor_layout, c->yCenter());
        addSpinBox(maskItem, m_editor_layout, c->xRadius());
        addSpinBox(maskItem, m_editor_layout, c->yRadius());
        // addSpinBox(maskItem, m_editor_layout, c->angle());
    } else if (auto* c = dynamic_cast<LineItem*>(maskItem))
        addSpinBox(maskItem, m_editor_layout, c->pos());
}

//! Private function, called by MaskPanel::setCurrentMaskItem
void MaskPanel::addMaskCheckBox(const QString& title, std::function<bool()> getter,
                                std::function<void(bool)> setter)
{
    auto* check_box = new QCheckBox(title, m_editor_layout->parentWidget());
    check_box->setChecked(getter());
    connect(check_box, &QCheckBox::stateChanged, [mask = m_current_mask_item, setter, check_box] {
        ASSERT(check_box);
        setter(check_box->isChecked());
        ASSERT(mask);
        emit mask->maskGeometryChanged();
        gDoc->setModified();
    });

    // TODO reconsider
    // connect(m_current_mask_item, &MaskItem::maskVisibilityChanged, [check_box, getter] {
    //     QSignalBlocker b(check_box);
    //     check_box->setChecked(getter());
    // });

    m_editor_layout->addRow(check_box);
}
